@@ -37,11 +37,11 @@ class gNMIclient(object):
     """
     This class instantiates the object, which interacts with the network elements over gNMI.
     """
-    def __init__(self, target: tuple, username: str = None, password: str = None,
+    def __init__(self, target: tuple, username: str = "", password: str = "",
                  debug: bool = False, insecure: bool = False, path_cert: str = None,
                  path_key: str = None, path_root: str = None, override: str = None,
-                 gnmi_timeout: int = 5, grpc_options: list = [], show_diff: str = "", 
-                 token: str = None, **kwargs):
+                 skip_verify = False, gnmi_timeout: int = 5, grpc_options: list = [],
+                 show_diff: str = "", token: str = None, **kwargs):
 
         """
         Initializing the object
@@ -53,10 +53,11 @@ class gNMIclient(object):
         self.__path_cert = path_cert
         self.__path_key = path_key
         self.__path_root = path_root
-        self.__options = ([('grpc.ssl_target_name_override', override)]+grpc_options) if override else grpc_options
+        self.__options = ([('grpc.ssl_target_name_override', override)] + grpc_options) if override else grpc_options
         self.__token = token
         self.__gnmi_timeout = gnmi_timeout
         self.__show_diff = show_diff if show_diff in {"get", "print"} else ""
+        self.__skip_verify = skip_verify
 
         self.__target_path = f'{target[0]}:{target[1]}'
         if re.match('unix:.*', target[0]):
@@ -111,21 +112,20 @@ class gNMIclient(object):
             # Certificates-based authentication
             if self.__path_cert and self.__path_key and self.__path_root:
                 try:
-                    cert = open(self.__path_cert, 'rb').read()
+                    ssl_cert = open(self.__path_cert, 'rb').read()
                     key = open(self.__path_key, 'rb').read()
                     root_cert = open(self.__path_root, 'rb').read()
-                    cert = grpc.ssl_channel_credentials(root_certificates=root_cert,
-                                                        private_key=key, certificate_chain=cert)
-                except:
+
+                except FileNotFoundError:
                     logging.error('The SSL certificate cannot be opened.')
                     raise Exception('The SSL certificate cannot be opened.')
 
             elif self.__path_cert:
                 try:
                     with open(self.__path_cert, 'rb') as f:
-                        cert = grpc.ssl_channel_credentials(f.read())
+                        ssl_cert = f.read()
 
-                except:
+                except FileNotFoundError:
                     logger.error('The SSL certificate cannot be opened.')
                     raise Exception('The SSL certificate cannot be opened.')
 
@@ -133,17 +133,43 @@ class gNMIclient(object):
             else:
                 try:
                     ssl_cert = ssl.get_server_certificate((self.__target[0], self.__target[1])).encode("utf-8")
-                    ssl_cert_deserialized = x509.load_pem_x509_certificate(ssl_cert, default_backend())
-                    ssl_cert_common_names = ssl_cert_deserialized.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)
-                    ssl_target_name_override = ssl_cert_common_names[0].value
-                    self.__options += [("grpc.ssl_target_name_override", ssl_target_name_override)]
-                    logger.warning('ssl_target_name_override is applied, should be used for testing only!')
-                    cert = grpc.ssl_channel_credentials(ssl_cert)
 
                 except:
                     logger.error(f'The SSL certificate cannot be retrieved from {self.__target}')
                     raise Exception(f'The SSL certificate cannot be retrieved from {self.__target}')
 
+            # Work with the certificate contents
+            ssl_cert_deserialized = x509.load_pem_x509_certificate(ssl_cert, default_backend())
+
+            # Collect Certificate's Common Name
+            ssl_cert_common_names = ssl_cert_deserialized.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)
+            ssl_target_name_override = ssl_cert_common_names[0].value
+
+            # Collect Certificate's Subject Alternative Names
+            self.__cert_sans = []
+            list_of_sans = [x509.IPAddress, x509.DNSName]
+            ssl_cert_subject_alt_names = ssl_cert_deserialized.extensions.get_extension_for_oid(x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
+
+            for entry in list_of_sans:
+                sans = ssl_cert_subject_alt_names.value.get_values_for_type(entry)
+                self.__cert_sans.extend([str(san) for san in sans])
+
+            # Set auto overrides
+            if self.__skip_verify:
+                self.__options.append(("grpc.ssl_target_name_override", ssl_target_name_override))
+                for san in self.__cert_sans:
+                    self.__options.append(("grpc.ssl_target_name_override", san))
+
+                logger.warning('ssl_target_name_override is applied, should be used for testing only!')
+
+            # Set up SSL channel credentials
+            if self.__path_key and self.__path_root:
+                cert = grpc.ssl_channel_credentials(root_certificates=root_cert,
+                                                    private_key=key, certificate_chain=ssl_cert)
+
+            else:
+                cert = grpc.ssl_channel_credentials(ssl_cert)
+
             # Build composed credentials if needed
             if self.__token:
                 call_credentials = grpc.access_token_call_credentials(access_token=self.__token)
@@ -240,14 +266,16 @@ class gNMIclient(object):
             print(f"Host: {self.__target_path}\nError: {err.details()}")
             logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
 
-            raise Exception (err)
+            raise Exception(err)
 
         except:
             logger.error('Collection of Capabilities is failed.')
 
             return None
 
-    def get(self, prefix: str = "", path: list = [], datatype: str = 'all', encoding: str = 'json'):
+    def get(self, prefix: str = "", path: list = [],
+            target: str = None, datatype: str = 'all',
+            encoding: str = 'json'):
         """
         Collecting the information about the resources from defined paths.
 
@@ -274,7 +302,7 @@ class gNMIclient(object):
           - ascii
           - json_ietf
         """
-        logger.info(f'Collecting info from requested paths (Get operation)...')
+        logger.info("Collecting info from requested paths (Get operation)...")
 
         datatype = datatype.lower()
         type_dict = {'all', 'config', 'state', 'operational'}
@@ -306,22 +334,24 @@ class gNMIclient(object):
 
         # Gnmi PREFIX
         try:
-            protobuf_prefix = gnmi_path_generator(prefix) if prefix else gnmi_path_generator([])
+            protobuf_prefix = gnmi_path_generator(prefix, target) if prefix else gnmi_path_generator([])
 
         except:
-            logger.error(f'Conversion of gNMI prefix to the Protobuf format failed')
-            raise Exception ('Conversion of gNMI prefix to the Protobuf format failed')
+            logger.error('Conversion of gNMI prefix to the Protobuf format failed')
+            raise Exception('Conversion of gNMI prefix to the Protobuf format failed')
 
         # Gnmi PATH
         try:
+            target = "" if prefix else target
+
             if not path:
                 protobuf_paths = []
-                protobuf_paths.append(gnmi_path_generator(path))
+                protobuf_paths.append(gnmi_path_generator(path, target))
             else:
-                protobuf_paths = [gnmi_path_generator(pe) for pe in path]
+                protobuf_paths = [gnmi_path_generator(pe, target) for pe in path]
 
         except:
-            logger.error(f'Conversion of gNMI paths to the Protobuf format failed')
+            logger.error('Conversion of gNMI paths to the Protobuf format failed')
             raise Exception ('Conversion of gNMI paths to the Protobuf format failed')
 
         if self.__capabilities and 'supported_encodings' in self.__capabilities:
@@ -356,29 +386,29 @@ class gNMIclient(object):
                     for notification in gnmi_message_response.notification:
                         notification_container = {}
 
-                        ## Message Notification, Key timestamp
+                        # Message Notification, Key timestamp
                         notification_container.update({'timestamp': notification.timestamp}) if notification.timestamp else notification_container.update({'timestamp': 0})
 
-                        ## Message Notification, Key prefix
+                        # Message Notification, Key prefix
                         notification_container.update({'prefix': gnmi_path_degenerator(notification.prefix)}) if notification.prefix else notification_container.update({'prefix': None})
 
-                        ## Message Notification, Key alias
+                        # Message Notification, Key alias
                         notification_container.update({'alias': notification.alias}) if notification.alias else notification_container.update({'alias': None})
 
-                        ## Message Notification, Key atomic
+                        # Message Notification, Key atomic
                         notification_container.update({'atomic': notification.atomic })
 
-                        ## Message Notification, Key update
+                        # Message Notification, Key update
                         if notification.update:
                             notification_container.update({'update': []})
 
                             for update_msg in notification.update:
                                 update_container = {}
 
-                                ## Message Update, Key path
+                                # Message Update, Key path
                                 update_container.update({'path': gnmi_path_degenerator(update_msg.path)}) if update_msg.path else update_container.update({'path': None})
 
-                                ## Message Update, Key val
+                                # Message Update, Key val
                                 if update_msg.HasField('val'):
                                     if update_msg.val.HasField('json_ietf_val'):
                                         update_container.update({'val': json.loads(update_msg.val.json_ietf_val)})
@@ -430,7 +460,9 @@ class gNMIclient(object):
 
             return None
 
-    def set(self, delete: list = None, replace: list = None, update: list = None, encoding: str = 'json'):
+    def set(self, delete: list = None, replace: list = None,
+            update: list = None, encoding: str = 'json',
+            target: str = None):
         """
         Changing the configuration on the destination network elements.
         Could provide a single attribute or multiple attributes.
@@ -466,7 +498,7 @@ class gNMIclient(object):
         if delete:
             if isinstance(delete, list):
                 try:
-                    del_protobuf_paths = [gnmi_path_generator(pe) for pe in delete]
+                    del_protobuf_paths = [gnmi_path_generator(pe, target) for pe in delete]
 
                 except:
                     logger.error(f'Conversion of gNMI paths to the Protobuf format failed')
@@ -480,7 +512,7 @@ class gNMIclient(object):
             if isinstance(replace, list):
                 for ue in replace:
                     if isinstance(ue, tuple):
-                        u_path = gnmi_path_generator(ue[0])
+                        u_path = gnmi_path_generator(ue[0], target)
                         u_val = json.dumps(ue[1]).encode('utf-8')
 
                         if encoding == 'json':
@@ -506,7 +538,7 @@ class gNMIclient(object):
             if isinstance(update, list):
                 for ue in update:
                     if isinstance(ue, tuple):
-                        u_path = gnmi_path_generator(ue[0])
+                        u_path = gnmi_path_generator(ue[0], target)
                         u_val = json.dumps(ue[1]).encode('utf-8')
 
                         if encoding == 'json':
@@ -632,7 +664,7 @@ class gNMIclient(object):
 
             raise rpc_ex
 
-    def _build_subscriptionrequest(self, subscribe: dict):
+    def _build_subscriptionrequest(self, subscribe: dict, target: str = None):
         if not isinstance(subscribe, dict):
             raise ValueError('Subscribe subscribe request is specified, but the value is not dict.')
 
@@ -707,17 +739,19 @@ class gNMIclient(object):
             subscribe.update({'prefix': None})
 
         if subscribe['prefix']:
-            request.prefix = gnmi_path_generator(subscribe['prefix'])
+            request.prefix = gnmi_path_generator(subscribe['prefix'], target)
 
         # subscription
         if 'subscription' not in subscribe or not subscribe['subscription']:
             raise ValueError('Subscribe:subscription value is missing.')
 
         for se in subscribe['subscription']:
+            # Amend target
+            target = "" if subscribe['prefix'] else target
             # path for that subscription
             if 'path' not in se:
                 raise ValueError('Subscribe:subscription:path is missing')
-            se_path = gnmi_path_generator(se['path'])
+            se_path = gnmi_path_generator(se['path'], target)
 
             # subscription entry mode; only relevent when the subscription request is stream
             if subscribe['mode'].lower() == 'stream':
@@ -793,7 +827,7 @@ class gNMIclient(object):
 
         return self.__stub.Subscribe(self.__generator(gnmi_message_request), metadata=self.__metadata)
 
-    def subscribe2(self, subscribe: dict):
+    def subscribe2(self, subscribe: dict, target: str = None):
         """
         New High-level method to serve temetry based on recent additions
         """
@@ -803,34 +837,33 @@ class gNMIclient(object):
 
         if subscribe['mode'].lower() in {'stream', 'once', 'poll'}:
             if subscribe['mode'].lower() == 'stream':
-                return self.subscribe_stream(subscribe=subscribe)
+                return self.subscribe_stream(subscribe=subscribe, target=target)
 
             elif subscribe['mode'].lower() == 'poll':
-                return self.subscribe_poll(subscribe=subscribe)
+                return self.subscribe_poll(subscribe=subscribe, target=target)
 
             elif subscribe['mode'].lower() == 'once':
-                return self.subscribe_once(subscribe=subscribe)
+                return self.subscribe_once(subscribe=subscribe, target=target)
 
         else:
             raise Exception('Unknown subscription request mode.')
 
-    def subscribe_stream(self, subscribe: dict):
+    def subscribe_stream(self, subscribe: dict, target: str = None):
         if 'mode' not in subscribe:
             subscribe['mode'] = 'STREAM'
-        gnmi_message_request = self._build_subscriptionrequest(subscribe)
+        gnmi_message_request = self._build_subscriptionrequest(subscribe, target)
         return StreamSubscriber(self.__channel, gnmi_message_request, self.__metadata)
 
-
-    def subscribe_poll(self, subscribe: dict):
+    def subscribe_poll(self, subscribe: dict, target: str = None):
         if 'mode' not in subscribe:
             subscribe['mode'] = 'POLL'
-        gnmi_message_request = self._build_subscriptionrequest(subscribe)
+        gnmi_message_request = self._build_subscriptionrequest(subscribe, target)
         return PollSubscriber(self.__channel, gnmi_message_request, self.__metadata)
 
-    def subscribe_once(self, subscribe: dict):
+    def subscribe_once(self, subscribe: dict, target: str = None):
         if 'mode' not in subscribe:
             subscribe['mode'] = 'ONCE'
-        gnmi_message_request = self._build_subscriptionrequest(subscribe)
+        gnmi_message_request = self._build_subscriptionrequest(subscribe, target)
         return OnceSubscriber(self.__channel, gnmi_message_request, self.__metadata)
 
     def __generator(self, in_message):
@@ -1069,24 +1102,8 @@ def telemetryParser(in_message=None, debug: bool = False):
             for update_msg in in_message.update.update:
                 update_container = {}
 
-                if update_msg.path and update_msg.path.elem:
-                    resource_path = []
-                    for path_elem in update_msg.path.elem:
-                        tp = ''
-                        if path_elem.name:
-                            tp += path_elem.name
-
-                        if path_elem.key:
-                            # Use 'sorted' to have a consistent ordering of keys
-                            for pk_name, pk_value in sorted(path_elem.key.items()):
-                                tp += f'[{pk_name}={pk_value}]'
-
-                        resource_path.append(tp)
-
-                    update_container.update({'path': '/'.join(resource_path)})
-
-                else:
-                    update_container.update({'path': None})
+                # Message Update, Key path
+                update_container.update({'path': gnmi_path_degenerator(update_msg.path)}) if update_msg.path else update_container.update({'path': None})
 
                 if update_msg.val:
                     if update_msg.val.HasField('json_ietf_val'):
